% Computes the equilibrium for a given parameter structure, p, and settings, s.
% Use baseline_joint_endogenous_settings.m to create default values of these parameters
function[g, w_l, w_h, gamma, F_l, F_h, F, f_l, f_h, f, distribution_shape, derivatives, grids, result, converged, resnorm] = solve_joint_endogenous(p, s)
    %% Default all to NaN in case it exits
    g=NaN; w_l=NaN; w_h=NaN; gamma=NaN; F_l=NaN; F_h=NaN; F=NaN; distribution_shape.EMPTY=NaN; derivatives.EMPTY=NaN; grids.EMPTY=NaN; f_l = NaN; f_h = NaN; f = NaN; resnorm = NaN;

    % Extract and check all parameters
    z_bar = p.z_bar;
    N = s.collocation_nodes;
    
    % Asserts on key parameter restrictions
    assert(p.z_bar > 0 && p.rho > 0 && p.lambda_l > 0 && p.lambda_h > 0 && p.Lambda >= 0 && p.zeta > 0 && p.psi > 0 && p.psi <= 1 && p.chi > 0, 'Parameters fail basic non-negativity constraints');    
        
    %% Sets up the chebyshev quadrature, etc.  Requires compecon: http://www4.ncsu.edu/~pfackler/compecon/toolbox.html
    % Calculates the basis and solves for the equations at the nodes and the boundaries
    polynomial_space = fundefn('cheb', N, 0, z_bar);
    z_vec_nodes = funnode(polynomial_space);
    z_vec = [0; z_vec_nodes; z_bar]; % i.e., including the boundaries   
    exp_z_vec = exp(z_vec);
    
    B = funbas(polynomial_space, z_vec, 0); % space for all F_l and F_h since shared collocation nodes
    B_p = funbas(polynomial_space, z_vec, 1); % basis for the derivatives
    [~, omega] = qnwcheb(N, 0, z_bar); % compute chebyshev quadrature weights.  These interior nodes line up with z_vec if the order 'N' matches
    omega = omega';
    
    %% Setup initial conditions by fitting polynomial coefficients to the provided functions
    if isfield(s, 'x_iv')
        x_iv = s.x_iv;
    else

        x_iv(1:N) = funfitf(polynomial_space, @(z) s.w_l_iv(z, p, s)); %finds chebyshev coefficients to approximate initial value
        x_iv(N+1:2*N) = funfitf(polynomial_space, @(z) s.w_h_iv(z, p, s));
        x_iv(2*N+1:3*N) = funfitf(polynomial_space, @(z) s.F_l_iv(z, p, s));
        x_iv(3*N+1:4*N) = funfitf(polynomial_space, @(z) s.F_h_iv(z, p, s));
        x_iv = x_iv'; %expects column vector
    end

    % Package up pre-calculated values in the settings for faster residual calculation
    s.B = B;
    s.B_p = B_p;
    s.omega = omega;
    s.exp_z_vec = exp_z_vec;
    s.z_vec = z_vec;
    s.N = N;

    % Matlab optimization options for the NLLS fit. Modify defaults 
    opts = optimoptions(@lsqnonlin,'Algorithm',s.Algorithm,'MaxIterations', s.max_iterations, 'FunctionTolerance', s.relative_tolerance, 'OptimalityTolerance', s.relative_tolerance, 'StepTolerance', s.relative_tolerance, 'Display', s.Display, 'UseParallel', s.UseParallel);

    % Solving system with nonlinear least-squares
    [x_sol,resnorm,residual,exitflag,result, lambda, jacobian] = lsqnonlin(@(x) joint_endogenous_L(x, p, s),x_iv,[],[],opts);
    
       
    if(resnorm > s.convergence_tolerance)
        if(s.debug_output)
            display_s('No convergence.  Last residual norm %d\n', resnorm);
        end
        converged = false;
        return;
    end
    
    % unpack the resulting coefficients
    grids.x_sol = x_sol;
    grids.c_l = x_sol(1:N);
    grids.c_h = x_sol(N+1:2*N);
    grids.d_l = x_sol(2*N+1:3*N);
    grids.d_h = x_sol(3*N+1:4*N);
    grids.B = B;
    grids.B_p = B_p;
    grids.z_vec = z_vec;

    % Use the basis and coefficients to go back to the function values
    grids.w_l_grid = B * grids.c_l;
    grids.w_h_grid = B * grids.c_h;
    grids.w_l_p_grid = B_p * grids.c_l;
    grids.w_h_p_grid = B_p * grids.c_h;
    grids.F_l_grid = B * grids.d_l;
    grids.F_h_grid = B * grids.d_h;
    grids.F_grid = grids.F_l_grid + grids.F_h_grid;
    grids.F_l_p_grid = B_p * grids.d_l;
    grids.F_h_p_grid = B_p * grids.d_h;
    grids.gamma_grid = grids.w_h_grid * p.chi/2;
    g = grids.gamma_grid(end);

    % Check that the F are monotone. If not, it did not convergence (and may not even have a solution for those parameters)
    F_diff_min = min(min(diff(grids.F_l_grid)), min(diff(grids.F_h_grid)));
    if(F_diff_min < s.F_monotonicity_threshold)
        if(s.debug_output)
            display_s('F not monotone.  Smallest diff %d\n', F_diff_min);
        end
        converged = false;
        return;        
    end    
   
    if(abs(grids.F_grid(end) - 1.0) > s.F_tolerance || (abs(grids.F_grid(1)) > s.F_tolerance))
        if(s.debug_output)
            display_s('No convergence.  Last residual norm %d, F(z_bar) = %d, F(0) = %d\n', resnorm, grids.F_grid(end), grids.F_grid(1));
        end
        converged = false;
        return;        
    end
   
    
    %% Prepare results by extracting from x, interpolating, etc.
    converged = true;

    %calculation of the Gini, etc.
    grids.alpha_grid =  (grids.F_h_p_grid + grids.F_l_p_grid)./(1 - grids.F_grid);
    distribution_shape.alpha_tail = grids.alpha_grid(floor(s.alpha_tail_proportion * length(grids.alpha_grid)), end);
    distribution_shape.F_0 =  grids.F_grid(1);
    distribution_shape.F_z_bar = grids.F_grid(end);
    distribution_shape.mean = omega * (1 - grids.F_grid(2:end-1)); %i.e., using chebyshev collocation with interior nodes
    distribution_shape.gini = 1 - omega * (1 - grids.F_grid(2:end-1)).^2 / distribution_shape.mean  ;
    
    
    if(distribution_shape.gini > 1 || distribution_shape.gini < 0)
        distribution_shape.gini = NaN;
    end
   
    % convenience utilities to interpolate results for plotting/etc.
    if(s.calculate_interpolation)
        w_l = s.interpolator(grids.z_vec, grids.w_l_grid);
        w_h = s.interpolator(grids.z_vec, grids.w_h_grid);
        gamma = s.interpolator(grids.z_vec, grids.gamma_grid);
        F_l = s.interpolator(grids.z_vec, grids.F_l_grid);
        F_h = s.interpolator(grids.z_vec, grids.F_h_grid);
        F = s.interpolator(grids.z_vec, grids.F_grid);
        f_l = s.interpolator(grids.z_vec, grids.F_l_p_grid);
        f_h = s.interpolator(grids.z_vec, grids.F_h_p_grid);
        f = s.interpolator(grids.z_vec, grids.F_h_p_grid + grids.F_l_p_grid);     
        derivatives.w_l_p = s.interpolator(grids.z_vec, grids.w_l_p_grid);
        derivatives.w_h_p = s.interpolator(grids.z_vec, grids.w_h_p_grid);   
        derivatives.F_p = s.interpolator(grids.z_vec, grids.F_h_p_grid + grids.F_l_p_grid);
        distribution_shape.alpha = s.interpolator(grids.z_vec, grids.alpha_grid);
        %For the tail parameter, finds the local parameter at some proportion of z_bar. Note: it becomes unstable at large z since F(z) near 0.
     end
 end